/* This module works with DT2821 card series from Data Translation */
/* In this version no DMA is used. I'm still fighting with it.     */
/* This driver comes WITHOUT ANY WARRENTY. It works for ME and does*/
/* not damage MY card. I'm not responsable for ANYTHING!           */
/* You can do with it what you want. Please contact me if you've   */
/* any problems with it.                                           */
/* BTW. It is configured for a card out of the box that means      */
/* Factory configuration                                           */
/*     Lakmal Gunasekara lucky@uni-bremen.de     4.6.1998          */
/*                                               Germany           */
/* TO DO                                                           */
/* Add gain                                                        */
/* Add more channels                                               */
/* Make it faster (use DMA) still developing                       */

/* version 0.1 */

#include <linux/errno.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <rtl.h>
#include <time.h>
#include <asm/io.h>

#include <pthread.h>

#include <rtl_sched.h>
#include <rtl_fifo.h>
#include "control.h"

pthread_t mytask;
int       sr;
char      channels;

/* for the DT2821 */

#define BIT0  0x0001
#define BIT1  0x0002
#define BIT2  0x0004
#define BIT3  0x0008
#define BIT4  0x0010
#define BIT5  0x0020
#define BIT6  0x0040
#define BIT7  0x0080
#define BIT8  0x0100
#define BIT9  0x0200
#define BIT10 0x0400
#define BIT11 0x0800
#define BIT12 0x1000
#define BIT13 0x2000
#define BIT14 0x4000
#define BIT15 0x8000

#define BASE    0x240               /* factory configuration */
#define ADCSR   BASE+0x0
#define CHANCSR BASE+0x2
#define ADDAT   BASE+0x4
#define DACSR   BASE+0x6
#define DADAT   BASE+0x8
#define DIODAT  BASE+0xA
#define SUPCSR  BASE+0xC
#define TMRCTR  BASE+0xE         

/* generates periodic output values         */
/* output range is 12 bit (0 to +4096)      */

int dt2821_getAD(int channel) {
       int data = 0;

       outw(BIT15, CHANCSR);                /* enable channel-gain list load */

       switch (channel) {                   /* choose Channel */
       case (0):
         outw(BIT9, ADCSR);
         break;
       case (1):
         outw(BIT9| BIT0, ADCSR);
         break;
       case (2):
         outw(BIT9| BIT1, ADCSR);
         break;
       case (3):
         outw (BIT9 | BIT1| BIT0, ADCSR);
         break;
       case (4):
         outw (BIT9 | BIT2, ADCSR);                   
         break;
       case (5):
         outw (BIT9 | BIT2 | BIT0, ADCSR);            
         break;
       case (6):
         outw (BIT9 | BIT2 | BIT1, ADCSR);            
         break;
       case (7):
         outw (BIT9 | BIT2 | BIT1 | BIT0, ADCSR);     
         break;     
       }
       
       outw(0, CHANCSR);                     /* end list loading */

       outw(BIT4, SUPCSR);                   /* preload multiplexer */

       while( inw(ADCSR) & BIT8) ;           /* wait until the MUX is ready */

       outw(BIT3, SUPCSR);                   /* software trigger */

       while(!(inw(ADCSR) & BIT7)) ;          /* wait until the conversion is do
ne */

       data=inw(ADDAT);

  return(data);         

}

/* .................................. */
/* function to write samples to fifos */
/* .................................. */
void *fun(void *t) {
  int data, rc;

  while(1) {
    /* FIFO's 0-7 transfer the samples from channel 1 to 8                       */
    /* FIFO 9 is an indicator whether a FIFO is full when trying to write to it  */
    if (channels==0) pthread_wait_np();
    if ((channels & 1)==1) {
      data=dt2821_getAD(0);
      if (rtf_put(0, (char *) &data, 2) < 0) {
        rc=0;
        rtf_put(9, (char *) &rc, sizeof(rc));
      }
      pthread_wait_np();
    }
    if ((channels & 2)==2) {
      data=dt2821_getAD(1);
      if (rtf_put(1, (char *) &data, 2) < 0) {
        rc=1;
        rtf_put(9, (char *) &rc, sizeof(rc));
      }
      pthread_wait_np();
    }
    if ((channels & 4)==4) {
      data=dt2821_getAD(2);
      if (rtf_put(2, (char *) &data, 2) < 0) {
        rc=2;
        rtf_put(9, (char *) &rc, sizeof(rc));
      }
      /* pthread_wait_np(); */
    }
    if ((channels & 8)==8) {
      data=dt2821_getAD(3);
      if (rtf_put(3, (char *) &data, 2) < 0) {
        rc=3;
        rtf_put(9, (char *) &rc, sizeof(rc));
      }
      /* pthread_wait_np(); */
    }
    if ((channels & 16)==16) {
      data=dt2821_getAD(4);
      if (rtf_put(4, (char *) &data, 2) < 0) {
        rc=4;
        rtf_put(9, (char *) &rc, sizeof(rc));
      }
      /* pthread_wait_np(); */
    }
    if ((channels & 32)==32) {
      data=dt2821_getAD(5);
      if (rtf_put(5, (char *) &data, 2) < 0) {
        rc=5;
        rtf_put(9, (char *) &rc, sizeof(rc));
      }
      /* pthread_wait_np(); */
    }
    if ((channels & 64)==64) {
      data=dt2821_getAD(6);
      if (rtf_put(6, (char *) &data, 2) < 0) {
        rc=6;
        rtf_put(9, (char *) &rc, sizeof(rc));
      }
      /* pthread_wait_np(); */
    }
    if ((channels & 128)==128) {
      data=dt2821_getAD(7);
      if (rtf_put(7, (char *) &data, 2) < 0) {
        rc=7;
        rtf_put(9, (char *) &rc, sizeof(rc));
      }
      /* pthread_wait_np(); */
    }
  }
}

int my_handler(unsigned int fifo) {
  struct my_msg_struct msg;
  int err;
  RTIME now;

  /* FIFO 8 transfers commands to control the driver */
  while ((err=rtf_get(8, &msg, sizeof(msg))) == sizeof(msg)) {
    switch(msg.command) {
    case START_TASK:
      now=gethrtime();
      pthread_make_periodic_np(mytask, now+NSECS_PER_SEC, HRTICKS_PER_SEC*(float)(1.0/(float)sr));
      pthread_wakeup_np(mytask);
      break;
    case STOP_TASK:
      pthread_suspend_np(mytask);
      break;
    case SET_CH:
      channels=(unsigned char)msg.value;
      break;
    case SET_POL:
      break;
    case SET_SR:
      sr=msg.value;
      break;
    default:
      return -EINVAL;
    }
  }
  if (err!=0) {
    return -EINVAL;
  }
  return 0;
}

int init_module(void) {
  pthread_attr_t     attr;
  struct sched_param sched_param;

  sr=1;
  channels=0x00;
  rtf_create(0,10000);
  rtf_create(1,10000);
  rtf_create(2,10000);
  rtf_create(3,10000);
  rtf_create(4,10000);
  rtf_create(5,10000);
  rtf_create(6,10000);
  rtf_create(7,10000);
  rtf_create(8,100);  /* input control channel   */
  rtf_create(9,100);  /* fifo overflow indicator */

  pthread_attr_init(&attr);
  sched_param.sched_priority = 2;
  pthread_attr_setschedparam(&attr, &sched_param);
  pthread_create(&mytask,  &attr, fun, (void *)1);
  pthread_setfp_np(mytask,1);
  rtf_create_handler(8, &my_handler);
  return 0;
}

void cleanup_module(void)
{
  rtf_destroy(0);
  rtf_destroy(1);
  rtf_destroy(2);
  rtf_destroy(3);
  rtf_destroy(4);
  rtf_destroy(5);
  rtf_destroy(6);
  rtf_destroy(7);
  rtf_destroy(8);
  rtf_destroy(9);
  pthread_delete_np(mytask);
}
